import traceback
from collections import defaultdict

from sqlalchemy import null, or_, select
from sqlalchemy.orm import joinedload

from chellow.dloads import put_mem_val
from chellow.e.computer import (
    SiteSource,
    SupplySource,
    contract_func,
    displaced_era,
    forecast_date,
)
from chellow.models import Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import c_months_u, ct_datetime_now, hh_format, hh_max, hh_min

CATS = ("imp_grid", "exp_grid", "used", "displaced")
QUANTS = (
    "kwh",
    "net_gbp",
    #    "p_per_kwh",
    "triad_gbp",
    "duos_gbp",
    "red_duos_gbp",
    "md_kw",
)


def _process(sess, log, start_year, start_month, mem_id, jval, site):
    log("started thread")
    f_date = forecast_date()
    report_context = {}

    for month_num, (month_start, month_finish) in enumerate(
        c_months_u(start_year=start_year, start_month=start_month, months=12)
    ):
        jval["progress"] = month_num
        put_mem_val(mem_id, jval)
        vals = {cat: defaultdict(int) for cat in CATS}
        vals["problem"] = ""

        for era in sess.execute(
            select(Era)
            .join(SiteEra)
            .join(Supply)
            .join(Source)
            .where(
                SiteEra.site == site,
                Era.start_date <= month_finish,
                or_(Era.finish_date == null(), Era.finish_date >= month_start),
                Era.imp_mpan_core != null(),
                Source.code.in_(("grid", "gen-grid")),
            )
            .options(
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.mop_contract),
                joinedload(Era.dc_contract),
            )
        ).scalars():
            chunk_start = hh_max(era.start_date, month_start)
            chunk_finish = hh_min(era.finish_date, month_finish)

            log("about to call first supplysource")
            supply_ds = SupplySource(
                sess, chunk_start, chunk_finish, f_date, era, True, report_context
            )

            supplier_contract = era.imp_supplier_contract
            import_vb_function = contract_func(
                report_context, supplier_contract, "virtual_bill"
            )
            if import_vb_function is None:
                raise Exception(
                    "Can't find the import_virtual_bill function in the "
                    "supplier contract. "
                )
            else:
                import_vb_function(supply_ds)
                v_bill = supply_ds.supplier_bill

                if "problem" in v_bill and len(v_bill["problem"]) > 0:
                    raise Exception("Supplier Problem: " + v_bill["problem"])

                try:
                    vals["imp_grid"]["net_gbp"] += v_bill["net-gbp"]
                    if "triad-gbp" in v_bill:
                        vals["imp_grid"]["triad_gbp"] += v_bill["triad-gbp"]
                    if "duos-red-gbp" in v_bill:
                        vals["imp_grid"]["red_duos_gbp"] += v_bill["duos-red-gbp"]
                    vals["imp_grid"]["duos_gbp"] += sum(
                        v
                        for k, v in v_bill.items()
                        if k.endswith("-gbp") and k.startswith("duos-")
                    )
                except KeyError:
                    raise Exception(
                        f"For the era {era.id} the virtual bill {v_bill} "
                        f"from the contract {supplier_contract} does not "
                        f"contain the net-gbp key."
                    )

            dc_contract = era.dc_contract
            supply_ds.contract_func(dc_contract, "virtual_bill")(supply_ds)
            dc_bill = supply_ds.dc_bill
            vals["imp_grid"]["gbp"] += dc_bill["net-gbp"]
            if "problem" in dc_bill and len(dc_bill["problem"]) > 0:
                raise Exception(f'DC Problem: {dc_bill["problem"]}')

            mop_contract = era.mop_contract
            mop_bill_function = supply_ds.contract_func(mop_contract, "virtual_bill")
            mop_bill_function(supply_ds)
            mop_bill = supply_ds.mop_bill
            vals["imp_grid"]["gbp"] += mop_bill["net-gbp"]
            if "problem" in mop_bill and len(mop_bill["problem"]) > 0:
                raise Exception(f'MOP Problem: {mop_bill["problem"]}')

            exp_supplier_contract = era.exp_supplier_contract
            if exp_supplier_contract is not None:
                log("about to call second supply source")
                exp_supply_ds = SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    f_date,
                    era,
                    False,
                    report_context,
                )
                exp_vb_function = contract_func(
                    report_context, exp_supplier_contract, "virtual_bill"
                )
                if exp_vb_function is None:
                    raise Exception(
                        "Can't find the export_virtual_bill function in the "
                        "supplier contract. "
                    )
                else:
                    exp_vb_function(exp_supply_ds)
                    v_bill = exp_supply_ds.supplier_bill

                    if "problem" in v_bill and len(v_bill["problem"]) > 0:
                        raise Exception(f"Export supplier Problem: {v_bill['problem']}")

                    try:
                        vals["exp_grid"]["net_gbp"] += v_bill["net-gbp"]
                        if "triad-gbp" in v_bill:
                            vals["exp_grid"]["triad_gbp"] += v_bill["triad-gbp"]
                        if "duos-red-gbp" in v_bill:
                            vals["exp_grid"]["red_duos_gbp"] += v_bill["duos-red-gbp"]
                        vals["exp_grid"]["duos_gbp"] += sum(
                            v
                            for k, v in v_bill.items()
                            if k.endswith("-gbp") and k.startswith("duos-")
                        )
                    except KeyError as e:
                        raise Exception(
                            f"For the era {era.id} the virtual bill {v_bill} from the "
                            f"contract {supplier_contract} does not contain the key {e}"
                        )

        disp_era = displaced_era(
            sess, report_context, site, month_start, month_finish, f_date
        )
        log("about to call sitesource")
        site_ds = SiteSource(
            sess,
            site,
            month_start,
            month_finish,
            f_date,
            report_context,
            disp_era,
            exclude_virtual=False,
        )

        if disp_era is not None:
            disp_supplier_contract = disp_era.imp_supplier_contract
            disp_vb_function = contract_func(
                report_context, disp_supplier_contract, "displaced_virtual_bill"
            )
            if disp_vb_function is None:
                raise Exception(
                    f"The supplier contract {disp_supplier_contract} "
                    "doesn't have the displaced_virtual_bill() function."
                )
            disp_vb_function(site_ds)
            disp_supplier_bill = site_ds.supplier_bill

            try:
                vals["displaced"]["net_gbp"] += disp_supplier_bill["net-gbp"]
            except KeyError:
                disp_supplier_bill["problem"] += (
                    f"For the supply {site_ds.mpan_core} the virtual bill "
                    f"{disp_supplier_bill} from the contract {disp_supplier_contract} "
                    f"does not contain the net-gbp key."
                )
            log(str(disp_supplier_bill))
            if "triad-gbp" in disp_supplier_bill:
                vals["displaced"]["triad_gbp"] += disp_supplier_bill["triad-gbp"]
            if "duos-red-gbp" in disp_supplier_bill:
                vals["displaced"]["red_duos_gbp"] += disp_supplier_bill["duos-red-gbp"]
            vals["displaced"]["duos_gbp"] += sum(
                v
                for k, v in disp_supplier_bill.items()
                if k.endswith("-gbp") and k.startswith("duos-")
            )

        for hh in site.hh_data(sess, month_start, month_finish, exclude_virtual=True):
            for cat in CATS:
                vals_cat = vals[cat]
                vals_cat["md_kw"] = max(vals_cat["md_kw"], hh[cat] * 2)
                vals_cat["kwh"] += float(hh[cat])

        for quant in ("net_gbp", "triad_gbp", "duos_gbp", "red_duos_gbp"):
            vals["used"][quant] += vals["imp_grid"][quant] + vals["displaced"][quant]

        """
        for cat in CATS:
            vals_cat = vals[cat]
            if vals_cat["kwh"] > 0:
                vals_cat["p_per_kwh"] = vals_cat["gbp"] / vals_cat["kwh"] * 100
            else:
                vals_cat["p_per_kwh"] = 0
        """

        FORMAT_LOOKUP = {
            "kwh": "{:,.0f}",
            "net_gbp": "{:,.2f}",
            "p_per_kwh": "{:,.2f}",
            "triad_gbp": "{:,.2f}",
            "duos_gbp": "{:,.2f}",
            "red_duos_gbp": "{:,.2f}",
            "md_kw": "{:,.2f}",
        }

        for quant in QUANTS:
            row = [hh_format(month_start)[:7], quant]
            fmt = FORMAT_LOOKUP[quant]
            for cat in CATS:
                row.append(fmt.format(vals[cat][quant]))
            jval["rows"].append(row)


def totals_runner(mem_id, site_id, start_date):
    sess = None
    titles = ["month", "quantity"]
    titles.extend(CATS)
    log_list = []

    jval = {"titles": titles, "rows": [], "status": "running", "log": log_list}
    try:
        with Session() as sess:
            site = Site.get_by_id(sess, site_id)

            def log(msg):
                return
                log_list.append(msg)
                put_mem_val(mem_id, jval)

            if start_date is None:
                now = ct_datetime_now()
                start_year = now.year - 1
                start_month = now.month
            else:
                start_year = start_date.year
                start_month = start_date.month

            _process(sess, log, start_year, start_month, mem_id, jval, site)

    except BaseException:
        msg = traceback.format_exc()
        jval["problem"] = msg
    finally:
        jval["status"] = "finished"
        put_mem_val(mem_id, jval)



import chellow.dloads
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_241 import content
from chellow.utils import ct_datetime, to_utc


def test_content(sess):
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    site = Site.insert(sess, "CI017", "Water Works 1")
    vf = to_utc(ct_datetime(2000, 1, 1))

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)

    mop_charge_script = """
from chellow.utils import reduce_bill_hhs

def virtual_bill_titles():
    return ['net-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]

        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))

    ds.mop_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, mop_charge_script, {}, vf, None, {}
    )

    dc_charge_script = """
from chellow.utils import reduce_bill_hhs

def virtual_bill_titles():
    return ['net-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]

        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))

    ds.dc_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, dc_charge_script, {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill_titles():
    return [
        'ccl-kwh', 'ccl-rate', 'ccl-gbp', 'net-gbp', 'vat-gbp', 'gross-gbp',
        'sum-msp-kwh', 'sum-msp-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * 0.1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        vf,
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        to_utc(ct_datetime(2022, 1, 1)),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()

    supply_id = supply.id
    start_year = 2022
    start_month = 1
    start_day = 1
    finish_year = 2022
    finish_month = 1
    finish_day = 31
    is_import = True
    content(
        start_year,
        start_month,
        start_day,
        finish_year,
        finish_month,
        finish_day,
        is_import,
        supply_id,
        user_id,
    )

    files = list(p.name for p in chellow.dloads.download_path.iterdir())
    assert files == ["00000_FINISHED_admin_example_com_daily_supplier_virtual_bill.csv"]



import csv
from io import StringIO

from utils import match, match_tables

from chellow.models import Site, User, UserRole
from chellow.reports.report_csv_site_hh_data import content
from chellow.utils import ct_datetime, to_utc


def test_do_get(client, mocker, sess):
    MockThread = mocker.patch(
        "chellow.reports.report_csv_site_hh_data.threading.Thread"
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))
    site_id = None
    exclude_virtual = False
    expected_args = start_date, finish_date, site_id, user.id, exclude_virtual
    query_string = {
        "start_year": "2020",
        "start_month": "01",
        "start_day": "01",
        "start_hour": "00",
        "start_minute": "00",
        "finish_year": "2020",
        "finish_month": "01",
        "finish_day": "31",
        "finish_hour": "23",
        "finish_minute": "30",
    }
    response = client.get("/reports/csv_site_hh_data", query_string=query_string)
    match(response, 303)
    MockThread.assert_called_with(target=content, args=expected_args)


def test_content(mocker, sess):
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch(
        "chellow.reports.report_csv_site_hh_data.open_file", return_value=mock_file
    )

    site = Site.insert(sess, "CI017", "Water Works")
    site_id = site.id
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id

    sess.commit()

    start_date = to_utc(ct_datetime(2024, 6, 1))
    finish_date = to_utc(ct_datetime(2024, 6, 1))
    exclude_virtual = False
    content(start_date, finish_date, site_id, user_id, exclude_virtual)
    mock_file.seek(0)
    actual = list(csv.reader(mock_file))
    expected = [
        [
            "site_id",
            "site_name",
            "associated_site_ids",
            "sources",
            "generator_types",
            "hh_start_clock_time",
            "imported_kwh",
            "displaced_kwh",
            "exported_kwh",
            "used_kwh",
            "parasitic_kwh",
            "generated_kwh",
            "3rd_party_import",
            "3rd_party_export",
            "meter_type",
        ],
        [
            "CI017",
            "Water Works",
            "",
            "",
            "",
            "2024-06-01 00:00",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "",
        ],
    ]

    match_tables(expected, actual)



import chellow.dloads
from chellow.e.computer import forecast_date
from chellow.models import Site, User, UserRole
from chellow.reports.report_181 import _write_sites, content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(sess):
    site = Site.insert(sess, "a", "a")
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    sess.commit()

    year = 2022
    site_id = site.id
    content(year, site_id, user_id)

    files = list(p.name for p in chellow.dloads.download_path.iterdir())
    assert files == ["00000_FINISHED_admin_example_com_output.csv"]


def test_write_sites(mocker):
    sess = mocker.Mock()
    caches = {}
    writer = mocker.Mock()
    year = 2010
    year_start = to_utc(ct_datetime(year - 1, 4, 1))
    year_finish = to_utc(ct_datetime(year, 3, 31, 23, 30))
    month_start = to_utc(utc_datetime(year, 3, 1))

    site_id = None
    site = mocker.Mock()
    era = mocker.Mock()
    source_codes = ("gen", "gen-net")

    ms = mocker.patch("chellow.reports.report_181._make_sites", autospec=True)
    ms.return_value = [site]
    de = mocker.patch("chellow.e.computer.displaced_era", autospec=True)
    de.return_value = era
    ss = mocker.patch("chellow.e.computer.SiteSource", autospec=True)
    ss_instance = ss.return_value
    ss_instance.hh_data = [
        {
            "start-date": month_start,
            "triad-actual-1-date": utc_datetime(year, 1, 12),
            "triad-actual-1-msp-kw": 12,
            "triad-actual-1-laf": 1,
            "triad-actual-1-gsp-kw": 12,
            "triad-actual-2-date": utc_datetime(year, 1, 12),
            "triad-actual-2-msp-kw": 12,
            "triad-actual-2-laf": 1,
            "triad-actual-2-gsp-kw": 12,
            "triad-actual-3-date": utc_datetime(year, 1, 12),
            "triad-actual-3-msp-kw": 12,
            "triad-actual-3-laf": 1,
            "triad-actual-3-gsp-kw": 12,
            "triad-actual-gsp-kw": 12,
            "triad-actual-rate": 10,
            "triad-actual-gbp": 120,
        }
    ]

    ss_instance.supplier_bill_hhs = {month_start: {}}

    mocker.patch("chellow.e.duos.duos_vb", autospec=True)
    mocker.patch("chellow.e.tnuos.hh", autospec=True)

    fd = forecast_date()

    _write_sites(sess, caches, writer, year, site_id)

    ms.assert_called_once_with(sess, year_start, year_finish, site_id, source_codes)
    de.assert_called_once_with(sess, caches, site, month_start, year_finish, fd)



from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from openpyxl import Workbook

from chellow.e.bill_parsers.settlement_dc_stark_xlsx import (
    Parser,
    _process_row,
    get_ct_date,
    get_gbp,
    make_column_lookup,
)
from chellow.utils import ct_datetime, to_ct, to_utc


def test_get_ct_date():
    dt_naive = Datetime(2024, 6, 30)

    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["E12"] = dt_naive
    dt = get_ct_date(sheet, "E", 12)

    assert dt == to_ct(dt_naive)


def test_get_gbp():
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["E12"] = None
    gbp = get_gbp(sheet, "E", 12)

    assert gbp is None


def test_process_row(sess):

    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["A6"] = "SSIL Billing Backing Data - Wessex"
    sheet["A7"] = "Date: 03/05/2024 09:16:58"

    sheet["G10"] = "Unmetered"
    sheet["J10"] = "Code 2"
    sheet["M10"] = "Code 3"
    sheet["P10"] = "Code 5"
    sheet["S10"] = "Code 10"
    sheet["V10"] = "GSM"
    sheet["Y10"] = "SavenergyOnline"
    sheet["AB10"] = "Meter Proving Tests"
    sheet["AE10"] = "Hand Held Visits (Ad hoc)"
    sheet["AH10"] = "Hand Held Visits (Regular)"
    sheet["AK10"] = "Annual Site Visits"

    sheet["A11"] = "Sage"
    sheet["B11"] = "MPAN"
    sheet["C11"] = "Site"
    sheet["D11"] = "Bill From"
    sheet["E11"] = "Bill To"
    sheet["F11"] = "Contract Reference"
    sheet["G11"] = "Unmetered Quant"
    sheet["H11"] = "Unmetered Rate"
    sheet["I11"] = "Quarterly Unmetered Charge"
    sheet["J11"] = "No. COP 2 Meters"
    sheet["K11"] = "Annual COP 2 DC/DA Rate"
    sheet["L11"] = "Quarterly COP 2 Charge"
    sheet["M11"] = "No. COP 3 Meters"
    sheet["N11"] = "Annual COP 3 DC/DA Rate"
    sheet["O11"] = "Quarterly COP 3 Charge"
    sheet["P11"] = "No. COP 5 Meters"
    sheet["Q11"] = "Annual COP 5 DC/DA Rate"
    sheet["R11"] = "Quarterly COP 5 Charge"
    sheet["S11"] = "No. COP 10 Meters"
    sheet["T11"] = "Annual COP 10 DC/DA Rate"
    sheet["U11"] = "Quarterly COP 10 Charge"
    sheet["V11"] = "No. GSM Meters"
    sheet["W11"] = "GSM Annual Rate"
    sheet["X11"] = "Quarterly GSM Charge"
    sheet["Y11"] = "No. Meters"
    sheet["Z11"] = "Annual SEO Rate"
    sheet["AA11"] = "Quarterly SEO Charge"
    sheet["AB11"] = "No. Meter Proving Test"
    sheet["AC11"] = "Meter Proving Test Rate"
    sheet["AD11"] = "Meter Proving Test Charge"
    sheet["AE11"] = "No. Hand Held Visits (Adhoc)"
    sheet["AF11"] = "Hand Held Visit (Adhoc) Rate"
    sheet["AG11"] = "Hand Held Visit (Adhoc) Charge"
    sheet["AH11"] = "No. Hand Held Visits (Regular)"
    sheet["AI11"] = "Hand Held Visit (Regular) Rate"
    sheet["AJ11"] = "Hand Held Visit (Regular) Charge"
    sheet["AK11"] = "No. Annual Site Visits"
    sheet["AL11"] = "Annual Site Visit Rate"
    sheet["AM11"] = "Annual Site Visit Charge"
    sheet["AN11"] = "Hand Held Visit Dates"
    sheet["AO11"] = "Grand Total"
    sheet["AP11"] = "VAT @ 20% "
    sheet["AQ11"] = "Grand Total"

    sheet["A12"] = "W006"
    sheet["B12"] = "2200013784589"
    sheet["C12"] = "NORTH PETHERTON STW, SOMERSET"
    sheet["D12"] = Datetime(2024, 4, 1)
    sheet["E12"] = Datetime(2024, 6, 30)
    sheet["F12"] = "GF002"
    sheet["G12"] = "0"
    sheet["H12"] = "0.00"
    sheet["I12"] = "0.00"
    sheet["J12"] = "0"
    sheet["K12"] = "0.00"
    sheet["L12"] = "0.00"
    sheet["M12"] = "0"
    sheet["N12"] = "99.50"
    sheet["O12"] = "0.00"
    sheet["P12"] = "1"
    sheet["Q12"] = "114.43"
    sheet["R12"] = "28.61"
    sheet["S12"] = "0"
    sheet["T12"] = "0.00"
    sheet["U12"] = "0.00"
    sheet["V12"] = "0"
    sheet["W12"] = "0.00"
    sheet["X12"] = "0.00"
    sheet["Y12"] = "1"
    sheet["Z12"] = "0.00"
    sheet["AA12"] = "0.00"
    sheet["AB12"] = "0"
    sheet["AC12"] = "25.00"
    sheet["AD12"] = "0.00"
    sheet["AE12"] = "0"
    sheet["AF12"] = "20.00"
    sheet["AG12"] = "0.00"
    sheet["AH12"] = "0"
    sheet["AI12"] = "0.00"
    sheet["AJ12"] = "0.00"
    sheet["AK12"] = "0"
    sheet["AL12"] = "20.00"
    sheet["AM12"] = "0.00"
    sheet["AN12"] = ""
    sheet["AO12"] = "28.61"
    sheet["AP12"] = "5.72"
    sheet["AQ12"] = "34.33"

    sheet["A13"] = ""  # Test copes with blank lines at the end

    row = 12
    issue_date = to_utc(ct_datetime(2025, 6, 2))
    column_lookup = make_column_lookup(sheet)
    bill = _process_row(sess, sheet, row, issue_date, column_lookup)

    assert bill == {
        "account": "22 0001 3784 589",
        "bill_type_code": "N",
        "breakdown": {
            "raw_lines": [],
            "settlement-status": [
                "settlement",
            ],
            "vat": {
                20: {
                    "net": Decimal("28.61"),
                    "vat": Decimal("5.72"),
                },
            },
        },
        "elements": [
            {
                "breakdown": {
                    "cop": {
                        "5",
                    },
                    "days": 91,
                    "mpan-days": 91,
                    "rate": {
                        Decimal("114.43"),
                    },
                },
                "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                "name": "mpan",
                "net": Decimal("28.61"),
                "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
        "gross": Decimal("34.33"),
        "issue_date": issue_date,
        "kwh": Decimal("0"),
        "mpan_core": "22 0001 3784 589",
        "net": Decimal("28.61"),
        "reads": [],
        "reference": "20240401_20240630_20250602_22 0001 3784 589",
        "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
        "vat": Decimal("5.72"),
    }


def test_process_row_element_sig_figs(sess):

    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["A6"] = "SSIL Billing Backing Data - Wessex"
    sheet["A7"] = "Date: 03/05/2024 09:16:58"

    sheet["G10"] = "Unmetered"
    sheet["J10"] = "Code 2"
    sheet["M10"] = "Code 3"
    sheet["P10"] = "Code 5"
    sheet["S10"] = "Code 10"
    sheet["V10"] = "GSM"
    sheet["Y10"] = "SavenergyOnline"
    sheet["AB10"] = "Meter Proving Tests"
    sheet["AE10"] = "Hand Held Visits (Ad hoc)"
    sheet["AH10"] = "Hand Held Visits (Regular)"
    sheet["AK10"] = "Annual Site Visits"

    sheet["A11"] = "Sage"
    sheet["B11"] = "MPAN"
    sheet["C11"] = "Site"
    sheet["D11"] = "Bill From"
    sheet["E11"] = "Bill To"
    sheet["F11"] = "Contract Reference"
    sheet["G11"] = "Unmetered Quant"
    sheet["H11"] = "Unmetered Rate"
    sheet["I11"] = "Quarterly Unmetered Charge"
    sheet["J11"] = "No. COP 2 Meters"
    sheet["K11"] = "Annual COP 2 DC/DA Rate"
    sheet["L11"] = "Quarterly COP 2 Charge"
    sheet["M11"] = "No. COP 3 Meters"
    sheet["N11"] = "Annual COP 3 DC/DA Rate"
    sheet["O11"] = "Quarterly COP 3 Charge"
    sheet["P11"] = "No. COP 5 Meters"
    sheet["Q11"] = "Annual COP 5 DC/DA Rate"
    sheet["R11"] = "Quarterly COP 5 Charge"
    sheet["S11"] = "No. COP 10 Meters"
    sheet["T11"] = "Annual COP 10 DC/DA Rate"
    sheet["U11"] = "Quarterly COP 10 Charge"
    sheet["V11"] = "No. GSM Meters"
    sheet["W11"] = "GSM Annual Rate"
    sheet["X11"] = "Quarterly GSM Charge"
    sheet["Y11"] = "No. Meters"
    sheet["Z11"] = "Annual SEO Rate"
    sheet["AA11"] = "Quarterly SEO Charge"
    sheet["AB11"] = "No. Meter Proving Test"
    sheet["AC11"] = "Meter Proving Test Rate"
    sheet["AD11"] = "Meter Proving Test Charge"
    sheet["AE11"] = "No. Hand Held Visits (Adhoc)"
    sheet["AF11"] = "Hand Held Visit (Adhoc) Rate"
    sheet["AG11"] = "Hand Held Visit (Adhoc) Charge"
    sheet["AH11"] = "No. Hand Held Visits (Regular)"
    sheet["AI11"] = "Hand Held Visit (Regular) Rate"
    sheet["AJ11"] = "Hand Held Visit (Regular) Charge"
    sheet["AK11"] = "No. Annual Site Visits"
    sheet["AL11"] = "Annual Site Visit Rate"
    sheet["AM11"] = "Annual Site Visit Charge"
    sheet["AN11"] = "Hand Held Visit Dates"
    sheet["AO11"] = "Grand Total"
    sheet["AP11"] = "VAT @ 20% "
    sheet["AQ11"] = "Grand Total"

    sheet["A12"] = "W006"
    sheet["B12"] = "2200013784589"
    sheet["C12"] = "NORTH PETHERTON STW, SOMERSET"
    sheet["D12"] = Datetime(2024, 4, 1)
    sheet["E12"] = Datetime(2024, 6, 30)
    sheet["F12"] = "GF002"
    sheet["G12"] = "0"
    sheet["H12"] = "0.00"
    sheet["I12"] = "0.00"
    sheet["J12"] = "0"
    sheet["K12"] = "0.00"
    sheet["L12"] = "0.00"
    sheet["M12"] = "0"
    sheet["N12"] = "99.50"
    sheet["O12"] = "0.00"
    sheet["P12"] = "1"
    sheet["Q12"] = "114.43"
    sheet["R12"] = "28.6"
    sheet["S12"] = "0"
    sheet["T12"] = "0.00"
    sheet["U12"] = "0.00"
    sheet["V12"] = "0"
    sheet["W12"] = "0.00"
    sheet["X12"] = "0.00"
    sheet["Y12"] = "1"
    sheet["Z12"] = "0.00"
    sheet["AA12"] = "0.00"
    sheet["AB12"] = "0"
    sheet["AC12"] = "25.00"
    sheet["AD12"] = "0.00"
    sheet["AE12"] = "0"
    sheet["AF12"] = "20.00"
    sheet["AG12"] = "0.00"
    sheet["AH12"] = "0"
    sheet["AI12"] = "0.00"
    sheet["AJ12"] = "0.00"
    sheet["AK12"] = "0"
    sheet["AL12"] = "20.00"
    sheet["AM12"] = "0.00"
    sheet["AN12"] = ""
    sheet["AO12"] = "28.61"
    sheet["AP12"] = "5.72"
    sheet["AQ12"] = "34.33"

    sheet["A13"] = ""  # Test copes with blank lines at the end

    row = 12
    issue_date = to_utc(ct_datetime(2025, 6, 2))
    column_lookup = make_column_lookup(sheet)
    bill = _process_row(sess, sheet, row, issue_date, column_lookup)

    assert bill == {
        "account": "22 0001 3784 589",
        "bill_type_code": "N",
        "breakdown": {
            "raw_lines": [],
            "settlement-status": [
                "settlement",
            ],
            "vat": {
                20: {
                    "net": Decimal("28.61"),
                    "vat": Decimal("5.72"),
                },
            },
        },
        "elements": [
            {
                "breakdown": {
                    "cop": {
                        "5",
                    },
                    "days": 91,
                    "mpan-days": 91,
                    "rate": {
                        Decimal("114.43"),
                    },
                },
                "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                "name": "mpan",
                "net": Decimal("28.60"),
                "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
            },
            {
                "breakdown": {
                    "activity-name": {
                        "discrepancy",
                    },
                },
                "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                "name": "activitiy",
                "net": Decimal("0.01"),
                "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
        "gross": Decimal("34.33"),
        "issue_date": issue_date,
        "kwh": Decimal("0"),
        "mpan_core": "22 0001 3784 589",
        "net": Decimal("28.61"),
        "reads": [],
        "reference": "20240401_20240630_20250602_22 0001 3784 589",
        "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
        "vat": Decimal("5.72"),
    }
    assert bill["elements"][0]["net"].as_tuple().exponent == -2


def test_process_row_old_format(sess):

    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["A6"] = "SSIL Billing Backing Data - Wessex"
    sheet["A7"] = "Date: 03/05/2024 09:16:58"

    sheet["G10"] = "Code 3"
    sheet["J10"] = "Code 5"
    sheet["M10"] = "GSM"
    sheet["P10"] = "SavenergyOnline"
    sheet["S10"] = "Meter Proving Tests"
    sheet["V10"] = "Hand Held Visits (Ad hoc)"
    sheet["Y10"] = "Hand Held Visits (Regular)"
    sheet["AB10"] = "Annual Site Visits"

    sheet["A11"] = "Sage"
    sheet["B11"] = "MPAN"
    sheet["C11"] = "Site"
    sheet["D11"] = "Bill From"
    sheet["E11"] = "Bill To"
    sheet["F11"] = "Contract Reference"
    sheet["G11"] = "No. COP 3 Meters"
    sheet["H11"] = "Annual COP 3 DC/DA Rate"
    sheet["I11"] = "COP 3 Charge"
    sheet["J11"] = "No. COP 5/10 Meters"
    sheet["K11"] = "Annual COP 5/10 DC/DA Rate"
    sheet["L11"] = "COP 5/10 Charge"
    sheet["M11"] = "No. GSM Meters"
    sheet["N11"] = "GSM Annual Rate"
    sheet["O11"] = "GSM Charge"
    sheet["P11"] = "No. Meters"
    sheet["Q11"] = "Annual SEO Rate"
    sheet["R11"] = "SEO Charge"
    sheet["S11"] = "No. Meter Proving Test"
    sheet["T11"] = "Meter Proving Test Rate"
    sheet["U11"] = "Meter Proving Test Charge"
    sheet["V11"] = "No. Hand Held Visits (Adhoc)"
    sheet["W11"] = "Hand Held Visit (Adhoc) Rate"
    sheet["X11"] = "Hand Held Visit (Adhoc) Charge"
    sheet["Y11"] = "No. Hand Held Visits (Regular)"
    sheet["Z11"] = "Hand Held Visit (Regular) Rate"
    sheet["AA11"] = "Hand Held Visit (Regular) Charge"
    sheet["AB11"] = "No. Annual Site Visits"
    sheet["AC11"] = "Annual Site Visit Rate"
    sheet["AD11"] = "Annual Site Visit Charge"
    sheet["AE11"] = "Hand Held Visit Dates"
    sheet["AF11"] = "Total Ex VAT"
    sheet["AG11"] = "VAT @ 20% "
    sheet["AH11"] = "Total In VAT"

    sheet["A12"] = "W006"
    sheet["B12"] = "2200013784589"
    sheet["C12"] = "NORTH PETHERTON STW, SOMERSET"
    sheet["D12"] = Datetime(2024, 4, 1)
    sheet["E12"] = Datetime(2024, 6, 30)
    sheet["F12"] = "GF002"
    sheet["G12"] = "0"
    sheet["H12"] = "99.50"
    sheet["I12"] = "0.00"
    sheet["J12"] = "1"
    sheet["K12"] = "114.43"
    sheet["L12"] = "28.61"
    sheet["M12"] = "0"
    sheet["N12"] = "0.00"
    sheet["O12"] = "0.00"
    sheet["P12"] = "1"
    sheet["Q12"] = "0.00"
    sheet["R12"] = "0.00"
    sheet["S12"] = "0"
    sheet["T12"] = "25.00"
    sheet["U12"] = "0.00"
    sheet["V12"] = "0"
    sheet["W12"] = "20.00"
    sheet["X12"] = "0.00"
    sheet["Y12"] = "0"
    sheet["Z12"] = "0.00"
    sheet["AA12"] = "0.00"
    sheet["AB12"] = "0"
    sheet["AC12"] = "20.00"
    sheet["AD12"] = "0.00"
    sheet["AE12"] = ""
    sheet["AF12"] = "28.61"
    sheet["AG12"] = "5.72"
    sheet["AH12"] = "34.33"

    sheet["A13"] = ""  # Test copes with blank lines at the end

    row = 12
    issue_date = to_utc(ct_datetime(2025, 6, 2))
    column_lookup = make_column_lookup(sheet)
    bill = _process_row(sess, sheet, row, issue_date, column_lookup)

    assert bill == {
        "account": "22 0001 3784 589",
        "bill_type_code": "N",
        "breakdown": {
            "raw_lines": [],
            "settlement-status": [
                "settlement",
            ],
            "vat": {
                20: {
                    "net": Decimal("28.61"),
                    "vat": Decimal("5.72"),
                },
            },
        },
        "elements": [
            {
                "breakdown": {
                    "cop": {
                        "5",
                    },
                    "days": 91,
                    "mpan-days": 91,
                    "rate": {
                        Decimal("114.43"),
                    },
                },
                "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                "name": "mpan",
                "net": Decimal("28.61"),
                "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
        "gross": Decimal("34.33"),
        "issue_date": issue_date,
        "kwh": Decimal("0"),
        "mpan_core": "22 0001 3784 589",
        "net": Decimal("28.61"),
        "reads": [],
        "reference": "20240401_20240630_20250602_22 0001 3784 589",
        "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
        "vat": Decimal("5.72"),
    }


def test_make_raw_bills(sess):

    f = BytesIO()
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 23)  # Have blank lines at the end
    sheet.insert_cols(0, 9)

    sheet["A6"] = "SSIL Billing Backing Data - Wessex"
    sheet["A7"] = "Date: 03/05/2024 09:16:58"

    sheet["G10"] = "Unmetered"
    sheet["J10"] = "Code 2"
    sheet["M10"] = "Code 3"
    sheet["P10"] = "Code 5"
    sheet["S10"] = "Code 10"
    sheet["V10"] = "GSM"
    sheet["Y10"] = "SavenergyOnline"
    sheet["AB10"] = "Meter Proving Tests"
    sheet["AE10"] = "Hand Held Visits (Ad hoc)"
    sheet["AH10"] = "Hand Held Visits (Regular)"
    sheet["AK10"] = "Annual Site Visits"

    sheet["A11"] = "Sage"
    sheet["B11"] = "MPAN"
    sheet["C11"] = "Site"
    sheet["D11"] = "Bill From"
    sheet["E11"] = "Bill To"
    sheet["F11"] = "Contract Reference"
    sheet["G11"] = "Unmetered Quant"
    sheet["H11"] = "Unmetered Rate"
    sheet["I11"] = "Quarterly Unmetered Charge"
    sheet["J11"] = "No. COP 2 Meters"
    sheet["K11"] = "Annual COP 2 DC/DA Rate"
    sheet["L11"] = "Quarterly COP 2 Charge"
    sheet["M11"] = "No. COP 3 Meters"
    sheet["N11"] = "Annual COP 3 DC/DA Rate"
    sheet["O11"] = "Quarterly COP 3 Charge"
    sheet["P11"] = "No. COP 5 Meters"
    sheet["Q11"] = "Annual COP 5 DC/DA Rate"
    sheet["R11"] = "Quarterly COP 5 Charge"
    sheet["S11"] = "No. COP 10 Meters"
    sheet["T11"] = "Annual COP 10 DC/DA Rate"
    sheet["U11"] = "Quarterly COP 10 Charge"
    sheet["V11"] = "No. GSM Meters"
    sheet["W11"] = "GSM Annual Rate"
    sheet["X11"] = "Quarterly GSM Charge"
    sheet["Y11"] = "No. Meters"
    sheet["Z11"] = "Annual SEO Rate"
    sheet["AA11"] = "Quarterly SEO Charge"
    sheet["AB11"] = "No. Meter Proving Test"
    sheet["AC11"] = "Meter Proving Test Rate"
    sheet["AD11"] = "Meter Proving Test Charge"
    sheet["AE11"] = "No. Hand Held Visits (Adhoc)"
    sheet["AF11"] = "Hand Held Visit (Adhoc) Rate"
    sheet["AG11"] = "Hand Held Visit (Adhoc) Charge"
    sheet["AH11"] = "No. Hand Held Visits (Regular)"
    sheet["AI11"] = "Hand Held Visit (Regular) Rate"
    sheet["AJ11"] = "Hand Held Visit (Regular) Charge"
    sheet["AK11"] = "No. Annual Site Visits"
    sheet["AL11"] = "Annual Site Visit Rate"
    sheet["AM11"] = "Annual Site Visit Charge"
    sheet["AN11"] = "Hand Held Visit Dates"
    sheet["AO11"] = "Grand Total"
    sheet["AP11"] = "VAT @ 20% "
    sheet["AQ11"] = "Grand Total"

    sheet["A12"] = "W006"
    sheet["B12"] = "2200013784589"
    sheet["C12"] = (
        "NORTH PETHERTON STW, PARKERS FIELD, NORTH PETHERTON, BRIDGWATER, SOMERSET"
    )
    sheet["D12"] = Datetime(2024, 4, 1)
    sheet["E12"] = Datetime(2024, 6, 30)
    sheet["F12"] = "GF002"
    sheet["G12"] = "0"
    sheet["H12"] = "0.00"
    sheet["I12"] = "0.00"
    sheet["J12"] = "0"
    sheet["K12"] = "0.00"
    sheet["L12"] = "0.00"
    sheet["M12"] = "0"
    sheet["N12"] = "99.50"
    sheet["O12"] = "0.00"
    sheet["P12"] = "1"
    sheet["Q12"] = "114.43"
    sheet["R12"] = "28.61"
    sheet["S12"] = "0"
    sheet["T12"] = "0.00"
    sheet["U12"] = "0.00"
    sheet["V12"] = "0"
    sheet["W12"] = "0.00"
    sheet["X12"] = "0.00"
    sheet["Y12"] = "1"
    sheet["Z12"] = "0.00"
    sheet["AA12"] = "0.00"
    sheet["AB12"] = "0"
    sheet["AC12"] = "25.00"
    sheet["AD12"] = "0.00"
    sheet["AE12"] = "0"
    sheet["AF12"] = "20.00"
    sheet["AG12"] = "0.00"
    sheet["AH12"] = "0"
    sheet["AI12"] = "0.00"
    sheet["AJ12"] = "0.00"
    sheet["AK12"] = "0"
    sheet["AL12"] = "20.00"
    sheet["AM12"] = "0.00"
    sheet["AN12"] = ""
    sheet["AO12"] = "28.61"
    sheet["AP12"] = "5.72"
    sheet["AQ12"] = "34.33"

    sheet["A13"] = "W006"
    sheet["B13"] = "1470001573345"
    sheet["C13"] = (
        "WICKWAR SEWAGE WORKS, OFF STATION ROAD, WICKWAR, WOTTON-UNDER-EDGE, "
        "GLOUCESTERSHIRE"
    )
    sheet["D13"] = Datetime(2024, 6, 12)
    sheet["E13"] = Datetime(2024, 6, 30)
    sheet["F13"] = "GF002"
    sheet["G13"] = "0"
    sheet["H13"] = "0.00"
    sheet["I13"] = "0.00"
    sheet["J13"] = "0"
    sheet["K13"] = "0.00"
    sheet["L13"] = "0.00"
    sheet["M13"] = "0"
    sheet["N13"] = "99.50"
    sheet["O13"] = "0.00"
    sheet["P13"] = "1"
    sheet["Q13"] = "114.43"
    sheet["R13"] = "5.96"
    sheet["S13"] = "0"
    sheet["T13"] = "0.00"
    sheet["U13"] = "0.00"
    sheet["V13"] = "0"
    sheet["W13"] = "0.00"
    sheet["X13"] = "0.00"
    sheet["Y13"] = "0.6333"
    sheet["Z13"] = "0.00"
    sheet["AA13"] = "0.00"
    sheet["AB13"] = "0"
    sheet["AC13"] = "25.00"
    sheet["AD13"] = "0.00"
    sheet["AE13"] = "0"
    sheet["AF13"] = "20.00"
    sheet["AG13"] = "0.00"
    sheet["AH13"] = "0"
    sheet["AI13"] = "0.00"
    sheet["AJ13"] = "0.00"
    sheet["AK13"] = "0"
    sheet["AL13"] = "20.00"
    sheet["AM13"] = "0.00"
    sheet["AN13"] = ""
    sheet["AO13"] = "5.96"
    sheet["AP13"] = "1.19"
    sheet["AQ13"] = "7.15"

    sheet["A14"] = ""  # Test copes with blank lines at the end

    wb.save(f)
    f.seek(0)
    p = Parser(f)
    raw_bills = p.make_raw_bills()
    assert raw_bills == [
        {
            "account": "22 0001 3784 589",
            "bill_type_code": "N",
            "breakdown": {
                "raw_lines": [],
                "settlement-status": [
                    "settlement",
                ],
                "vat": {
                    20: {
                        "net": Decimal("28.61"),
                        "vat": Decimal("5.72"),
                    },
                },
            },
            "elements": [
                {
                    "breakdown": {
                        "cop": {
                            "5",
                        },
                        "days": 91,
                        "mpan-days": 91,
                        "rate": {
                            Decimal("114.43"),
                        },
                    },
                    "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                    "name": "mpan",
                    "net": Decimal("28.61"),
                    "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
                },
            ],
            "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
            "gross": Decimal("34.33"),
            "issue_date": to_utc(ct_datetime(2024, 5, 3, 9, 16)),
            "kwh": Decimal("0"),
            "mpan_core": "22 0001 3784 589",
            "net": Decimal("28.61"),
            "reads": [],
            "reference": "20240401_20240630_20240503_22 0001 3784 589",
            "start_date": to_utc(ct_datetime(2024, 4, 1, 0, 0)),
            "vat": Decimal("5.72"),
        },
        {
            "account": "14 7000 1573 345",
            "bill_type_code": "N",
            "breakdown": {
                "raw_lines": [],
                "settlement-status": [
                    "settlement",
                ],
                "vat": {
                    20: {
                        "net": Decimal("5.96"),
                        "vat": Decimal("1.19"),
                    },
                },
            },
            "elements": [
                {
                    "breakdown": {
                        "cop": {
                            "5",
                        },
                        "days": 19,
                        "mpan-days": 19,
                        "rate": {
                            Decimal("114.43"),
                        },
                    },
                    "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
                    "name": "mpan",
                    "net": Decimal("5.96"),
                    "start_date": to_utc(ct_datetime(2024, 6, 12, 0, 0)),
                },
            ],
            "finish_date": to_utc(ct_datetime(2024, 6, 30, 23, 30)),
            "gross": Decimal("7.15"),
            "issue_date": to_utc(ct_datetime(2024, 5, 3, 9, 16)),
            "kwh": Decimal("0"),
            "mpan_core": "14 7000 1573 345",
            "net": Decimal("5.96"),
            "reads": [],
            "reference": "20240612_20240630_20240503_14 7000 1573 345",
            "start_date": to_utc(ct_datetime(2024, 6, 12, 0, 0)),
            "vat": Decimal("1.19"),
        },
    ]



import atexit
import collections
import threading
import traceback

from sqlalchemy.sql.expression import select

from chellow.models import (
    Contract,
    GContract,
    RSession,
    Report,
)
from chellow.utils import ct_datetime_now

tester = None


def _run(log_f, sess):
    log_f("Starting to run tests.")
    for report in sess.execute(select(Report).order_by(Report.id)).scalars():
        _test_report(log_f, sess, report)
        sess.rollback()  # Avoid long-running transaction
    for contract in sess.execute(select(Contract).order_by(Contract.id)).scalars():
        _test_contract(log_f, sess, contract)
        sess.rollback()  # Avoid long-running transaction
    for g_contract in sess.execute(select(GContract).order_by(GContract.id)).scalars():
        _test_g_contract(log_f, sess, g_contract)
        sess.rollback()  # Avoid long-running transaction


class Tester(threading.Thread):
    def __init__(self):
        super().__init__(name="Tester")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=500)
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                self.global_alert = None
                with RSession() as sess:
                    try:
                        _run(self.log, sess)
                    except BaseException:
                        self.log(traceback.format_exc())
                        self.global_alert = (
                            "There's a problem with the "
                            "<a href='/tester'>Automatic Tester</a>."
                        )
                        sess.rollback()
                    finally:
                        self.lock.release()
                        self.log("Finished running tests.")

            self.going.wait(60 * 60 * 24)
            self.going.clear()


def _test_report(logger, sess, report):
    logger(f"Starting to test local report {report.id} {report.name}.")
    code = compile(report.script, "<string>", "exec")
    ns = {"report_id": report.id, "template": report.template}
    exec(code, ns)
    if "test" in ns:
        ns["test"]()


def _test_contract(logger, sess, contract):
    logger(
        f"Starting to test {contract.market_role.description} contract {contract.id} "
        f"{contract.name}."
    )
    code = compile(contract.charge_script, "<string>", "exec")
    ns = {"db_id": contract.id}
    exec(code, ns)
    if "test" in ns:
        ns["test"]()


def _test_g_contract(logger, sess, g_contract):
    logger(
        f"Starting to test gas {'industry' if g_contract.is_industry else 'supplier'} "
        f"contract {g_contract.id} {g_contract.name}."
    )
    code = compile(g_contract.charge_script, "<string>", "exec")
    ns = {"db_id": g_contract.id, "properties": g_contract.make_properties()}
    exec(code, ns)
    if "test" in ns:
        ns["test"]()


def get_importer():
    return tester


def startup():
    global tester
    tester = Tester()
    tester.start()


@atexit.register
def shutdown():
    if tester is not None:
        tester.stop()



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_mpan_core, to_ct, to_utc


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_ct_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if isinstance(val, Datetime):
        dt = val
    elif isinstance(val, str):
        dt = Datetime.strptime(val[:10], "%d/%m/%Y")
    else:
        raise BadRequest(
            f"The value {val} at {cell.coordinate} is of type {type(val)}, but "
            f"expected a timestamp or string."
        )
    return to_ct(dt)


def _process_row(issue_date, sheet, row):
    mpan_core = parse_mpan_core(str(get_int(sheet, "B", row)))

    start_date = finish_date = to_utc(get_ct_date(sheet, "F", row))
    activity_name_raw = get_str(sheet, "G", row)
    activity_name = activity_name_raw.lower().replace(" ", "_")

    net_dec = get_dec(sheet, "I", row)
    net = round(net_dec, 2)

    vat_dec = get_dec(sheet, "J", row)
    vat = round(vat_dec, 2)

    gross_dec = get_dec(sheet, "K", row)
    gross = round(gross_dec, 2)

    breakdown = {
        "raw-lines": [],
        "vat": {20: {"vat": vat, "net": net}},
    }

    return {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "vat": vat,
        "net": net,
        "gross": gross,
        "reads": [],
        "breakdown": breakdown,
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": mpan_core,
        "reference": "_".join(
            (
                start_date.strftime("%Y%m%d"),
                finish_date.strftime("%Y%m%d"),
                issue_date.strftime("%Y%m%d"),
                mpan_core,
            )
        ),
        "elements": [
            {
                "name": "activity",
                "start_date": start_date,
                "finish_date": finish_date,
                "net": net,
                "breakdown": {"name": {activity_name}},
            }
        ],
    }


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[1]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        try:
            bills = []
            issue_date = to_utc(get_ct_date(self.sheet, "C", 6))
            for row in range(12, len(self.sheet["A"]) + 1):
                val = get_cell(self.sheet, "B", row).value
                if val is None or val == "":
                    break
                self._set_last_line(row, val)

                bill = _process_row(issue_date, self.sheet, row)
                bills.append(bill)

        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills



from io import BytesIO
from utils import match


def test_edit_viewer_post(client, mocker):
    file_name = "sample.edi"
    file_bytes = "MHD=1+UTLHDR:3'".encode("utf8")
    f = BytesIO(file_bytes)

    data = {"edi_file": (f, file_name)}
    response = client.post("/edi_viewer", data=data)

    match(response, 200)



from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from openpyxl import Workbook

from chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx import Parser, _process_row
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_process_row(mocker, sess):
    rates = {
        "annual_rates": {
            "non_settlement": {"*": {"IP": {"*": {"gbp_per_meter": Decimal("12")}}}}
        }
    }
    mocker.patch(
        "chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx.hh_rate", return_value=rates
    )
    caches = {}
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 2)
    sheet.insert_cols(0, 9)

    sheet["A1"] = "NAME"
    sheet["B1"] = "NMR"
    sheet["C1"] = "METER TYPE"
    sheet["D1"] = "IDENTIFIER"
    sheet["E1"] = "METER"
    sheet["F1"] = "MPAN REF"
    sheet["G1"] = "START"
    sheet["H1"] = "END"
    sheet["I1"] = "CHECK"

    sheet["A2"] = "Treglisson"
    sheet["B2"] = "Q997"
    sheet["C2"] = "VIS"
    sheet["D2"] = 101
    sheet["E2"] = "88jiuf ff"
    sheet["F2"] = 1472066139971
    sheet["G2"] = Datetime(2018, 6, 1)
    sheet["H2"] = Datetime(2018, 6, 30)
    sheet["I2"] = "Billed"
    bill = _process_row(caches, sess, sheet[1], sheet[2])
    assert bill == {
        "account": "14 7206 6139 971",
        "bill_type_code": "N",
        "breakdown": {
            "cop": [
                "5",
            ],
            "msn": [
                "88jiuf ff",
            ],
            "raw_lines": [],
            "vat": {
                20: {
                    "net": Decimal("1.00"),
                    "vat": Decimal("0.20"),
                },
            },
        },
        "elements": [
            {
                "breakdown": {
                    "comm": {
                        "IP",
                    },
                    "months": 1,
                    "rate": {
                        Decimal("12"),
                    },
                    "settlement-status": {
                        "non_settlement",
                    },
                },
                "finish_date": to_utc(ct_datetime(2018, 6, 30, 23, 30)),
                "name": "meter",
                "net": Decimal("1.00"),
                "start_date": to_utc(ct_datetime(2018, 6, 1, 0, 0)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2018, 6, 30, 23, 30)),
        "gross": Decimal("1.20"),
        "issue_date": to_utc(ct_datetime(2018, 6, 1)),
        "kwh": Decimal("0"),
        "mpan_core": "14 7206 6139 971",
        "net": Decimal("1.00"),
        "reads": [],
        "reference": "20180601_20180630_20180601_14 7206 6139 971",
        "start_date": to_utc(ct_datetime(2018, 6, 1)),
        "vat": Decimal("0.20"),
    }


def test(mocker, sess):
    rates = {
        "annual_rates": {
            "non_settlement": {"*": {"IP": {"*": {"gbp_per_meter": Decimal("45")}}}}
        }
    }
    mocker.patch(
        "chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx.hh_rate", return_value=rates
    )
    f = BytesIO()
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 2)
    sheet.insert_cols(0, 9)

    sheet["A1"] = "NAME"
    sheet["B1"] = "NMR"
    sheet["C1"] = "METER TYPE"
    sheet["D1"] = "IDENTIFIER"
    sheet["E1"] = "METER"
    sheet["F1"] = "MPAN REF"
    sheet["G1"] = "START"
    sheet["H1"] = "END"
    sheet["I1"] = "CHECK"

    sheet["A2"] = "Treglisson"
    sheet["B2"] = "Q997"
    sheet["C2"] = "VIS"
    sheet["D2"] = 101
    sheet["E2"] = "88jiuf ff"
    sheet["F2"] = 1472066139971
    sheet["G2"] = Datetime(2018, 6, 1)
    sheet["H2"] = Datetime(2018, 6, 30)
    sheet["I2"] = "Billed"

    wb.save(f)
    f.seek(0)
    parser = Parser(f)
    parser.make_raw_bills()


def test_old(mocker, sess):
    rates = {
        "annual_rates": {
            "non_settlement": {"*": {"IP": {"*": {"gbp_per_meter": Decimal("45")}}}}
        }
    }
    mocker.patch(
        "chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx.hh_rate", return_value=rates
    )
    f = BytesIO()
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 2)
    sheet.insert_cols(0, 9)

    sheet["A1"] = "MPAN"
    sheet["B1"] = "WESSEX WATER SITE NAME"
    sheet["C1"] = "STARK IDENTIFIER"
    sheet["D1"] = "NMR"
    sheet["E1"] = "METER TYPE"
    sheet["F1"] = "IDENTIFIER"
    sheet["G1"] = "METER"
    sheet["H1"] = "MPAN REF"
    sheet["I1"] = "START"
    sheet["J1"] = "END"
    sheet["K1"] = "CHECK"

    sheet["A2"] = 1472066139971
    sheet["B2"] = ""
    sheet["C2"] = ""
    sheet["D2"] = "Q997"
    sheet["E2"] = "VIS"
    sheet["F2"] = 101
    sheet["G2"] = "88jiuf ff"
    sheet["H2"] = 1472066139971
    sheet["I2"] = Datetime(2018, 7, 1)
    sheet["J2"] = Datetime(2018, 7, 1)
    sheet["K2"] = "Billed"

    wb.save(f)
    f.seek(0)

    parser = Parser(f)
    parser.make_raw_bills()


def test_one_bill(mocker, sess):
    f = BytesIO()
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 2)
    sheet.insert_cols(0, 9)

    sheet["A1"] = "NAME"
    sheet["B1"] = "NMR"
    sheet["C1"] = "METER TYPE"
    sheet["D1"] = "IDENTIFIER"
    sheet["E1"] = "METER"
    sheet["F1"] = "MPAN REF"
    sheet["G1"] = "START"
    sheet["H1"] = "END"
    sheet["I1"] = "CHECK"

    sheet["A2"] = "Quantum Computer"
    sheet["B2"] = 65837766
    sheet["C2"] = "x600"
    sheet["D2"] = 7493477439
    sheet["E2"] = 7864739737
    sheet["F2"] = 2088757371777
    sheet["G2"] = Datetime(2022, 2, 1)
    sheet["H2"] = Datetime(2022, 2, 28)
    sheet["I2"] = "Billed"

    sheet["C6"].value = Datetime(2022, 3, 1)

    sheet["B12"].value = "2227651294270"
    sheet["C12"].value = "CS"
    sheet["D12"].value = "Settled"
    sheet["E12"].value = "Massive Field, Taunton, Somerset"
    sheet["F12"].value = Datetime(2022, 1, 1)
    sheet["G12"].value = Datetime(2022, 1, 31)
    sheet["H12"].value = "5.00"
    sheet["I12"].value = "0.12"
    sheet["J12"].value = "0.76"
    sheet["K12"].value = "0.99"

    wb.save(f)
    f.seek(0)
    rates = {
        "annual_rates": {
            "non_settlement": {"*": {"IP": {"*": {"gbp_per_meter": Decimal("45")}}}}
        }
    }
    mocker.patch(
        "chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx.hh_rate", return_value=rates
    )
    p = Parser(f)
    result = p.make_raw_bills()

    expected = [
        {
            "bill_type_code": "N",
            "kwh": Decimal("0"),
            "vat": Decimal("0.75"),
            "net": Decimal("3.75"),
            "gross": Decimal("4.50"),
            "reads": [],
            "breakdown": {
                "raw_lines": [],
                "cop": ["5"],
                "msn": ["7864739737"],
                "vat": {
                    20: {
                        "net": Decimal("3.75"),
                        "vat": Decimal("0.75"),
                    },
                },
            },
            "account": "20 8875 7371 777",
            "issue_date": utc_datetime(2022, 2, 1),
            "start_date": utc_datetime(2022, 2, 1),
            "finish_date": utc_datetime(2022, 2, 28, 23, 30),
            "mpan_core": "20 8875 7371 777",
            "reference": "20220201_20220228_20220201_20 8875 7371 777",
            "elements": [
                {
                    "name": "meter",
                    "start_date": utc_datetime(2022, 2, 1),
                    "finish_date": utc_datetime(2022, 2, 28, 23, 30),
                    "net": Decimal("3.75"),
                    "breakdown": {
                        "rate": {Decimal("45")},
                        "months": 1,
                        "settlement-status": {"non_settlement"},
                        "comm": {"IP"},
                    },
                }
            ],
        }
    ]

    assert result == expected



from decimal import Decimal, InvalidOperation

from openpyxl import load_workbook
from openpyxl.utils.datetime import from_excel

from werkzeug.exceptions import BadRequest

from chellow.e.computer import hh_rate
from chellow.models import Session
from chellow.utils import ct_datetime, parse_mpan_core, to_ct, to_utc


def get_ct_date(title_row, row, name):
    val_raw = get_value(title_row, row, name)
    if isinstance(val_raw, int):
        val = from_excel(val_raw)
    else:
        val = val_raw
    return to_ct(val)


def get_start_date(title_row, row, name):
    return to_utc(get_ct_date(title_row, row, name))


def get_value(title_row, row, name):
    idx = None
    name = name.strip().lower()

    for i, title_cell in enumerate(title_row):
        if str(title_cell.value).strip().lower() == name:
            idx = i
            break

    if idx is None:
        raise BadRequest(
            "The name '{name}' can't be found in the titles "
            "{title_row}.".format(name=name, title_row=title_row)
        )

    try:
        return row[idx].value

    except IndexError:
        raise BadRequest(
            f"For the row {row}, the index is {idx} which is beyond the end of the row."
        )


def get_str(title_row, row, name):
    return get_value(title_row, row, name).strip()


def get_dec(title_row, row, name):
    try:
        return Decimal(str(get_value(title_row, row, name)))
    except InvalidOperation:
        return None


def get_int(title_row, row, name):
    return int(get_value(title_row, row, name))


def _process_row(caches, sess, title_row, row):
    msn = str(get_value(title_row, row, "meter")).strip()
    mpan_core = parse_mpan_core(str(get_int(title_row, row, "mpan ref")))
    start_date_ct = get_ct_date(title_row, row, "start")
    start_date = to_utc(start_date_ct)
    issue_date = start_date
    finish_date_ct = get_ct_date(title_row, row, "end")
    finish_date = to_utc(
        ct_datetime(
            finish_date_ct.year, finish_date_ct.month, finish_date_ct.day, 23, 30
        )
    )
    rates = hh_rate(sess, caches, 0, start_date)
    meter_rate = rates["annual_rates"]["non_settlement"]["*"]["IP"]["*"][
        "gbp_per_meter"
    ]
    months = (finish_date_ct.year - start_date_ct.year) * 12 + (
        finish_date_ct.month - start_date_ct.month + 1
    )
    net = round(Decimal(float(meter_rate) / 12 * months), 2)
    vat = round(net * Decimal("0.2"), 2)
    elements = [
        {
            "name": "meter",
            "start_date": start_date,
            "finish_date": finish_date,
            "net": net,
            "breakdown": {
                "rate": {meter_rate},
                "months": months,
                "settlement-status": {"non_settlement"},
                "comm": {"IP"},
            },
        }
    ]

    breakdown = {
        "raw_lines": [],
        "cop": ["5"],
        "msn": [msn],
        "vat": {20: {"vat": vat, "net": net}},
    }

    return {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "vat": vat,
        "net": net,
        "gross": net + vat,
        "reads": [],
        "breakdown": breakdown,
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": mpan_core,
        "reference": "_".join(
            (
                start_date_ct.strftime("%Y%m%d"),
                finish_date_ct.strftime("%Y%m%d"),
                start_date_ct.strftime("%Y%m%d"),
                mpan_core,
            )
        ),
        "elements": elements,
    }


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row_index = None
        caches = {}
        try:
            with Session() as sess:
                bills = []
                row_iter = self.sheet.iter_rows()
                title_row = next(row_iter)
                for row in row_iter:
                    val = get_value(title_row, row, "mpan ref")
                    if val is None or val == "":
                        break
                    if get_str(title_row, row, "check") != "Billed":
                        continue

                    self._set_last_line(row_index, val)
                    bill = _process_row(caches, sess, title_row, row)
                    bills.append(bill)
                    sess.rollback()
        except BadRequest as e:
            raise BadRequest(f"Row number: {row_index} {e.description}")

        return bills



import json

from decimal import Decimal
from io import BytesIO

from sqlalchemy import select
from chellow.models import Contract, RateScript
from chellow.rate_server import download
from chellow.utils import ct_datetime, hh_format, to_utc


def hh(supply_source):
    try:
        ro_cache = supply_source.caches["ro"]
    except KeyError:
        ro_cache = supply_source.caches["ro"] = {}

    for hh in supply_source.hh_data:
        try:
            hh["ro-rate"] = ro_rate = ro_cache[hh["start-date"]]
        except KeyError:
            h_start = hh["start-date"]
            hh["ro-rate"] = ro_rate = ro_cache[h_start] = float(
                supply_source.non_core_rate("ro", h_start)["ro_gbp_per_msp_kwh"]
            )

        ro_kwh = hh["msp-kwh"]
        hh["ro-msp-kwh"] = ro_kwh
        hh["ro-kwh"] = ro_kwh
        hh["ro-gbp"] = ro_kwh * ro_rate


def rate_server_import(sess, log, set_progress, paths):
    log("Starting to check for new electricity RO rates")
    contract_name = "ro"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    contract_props = contract.make_properties()

    if not contract_props.get("enabled", True):
        log(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to true."
        )
        return

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year, utility, rate_type, file_name = path
            if utility == "electricity" and rate_type == "ro":
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_files in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        if year_start < contract.start_rate_script.start_date:
            continue
        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, year_start, {})

        if len(year_files) > 0:
            file_name, url = sorted(year_files.items())[-1]

            rs_script = rs.make_script()
            if rs_script.get("a_file_name") != file_name:
                log(
                    f"Found new file {file_name} for rate script starting "
                    f"{hh_format(year_start)}"
                )
                script = {"a_file_name": file_name}
                server_j = json.load(BytesIO(download(url)))
                script["ro_gbp_per_msp_kwh"] = Decimal(server_j["ro_gbp_per_msp_kwh"])
                rs.update(script)
                log(f"Updated RO rate script for {hh_format(year_start)}")

    log("Finished RO rates")
    sess.commit()



import datetime
import traceback
from datetime import timedelta as Timedelta

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

import xlrd

from zish import loads

from chellow.e.elexon import download_file
from chellow.models import Contract, RateScript
from chellow.utils import HH, ct_datetime, hh_format, to_ct, to_utc


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source):
    for h in data_source.hh_data:
        try:
            sbp, ssp = data_source.caches["system_price"][h["start-date"]]
        except KeyError:
            try:
                system_price_cache = data_source.caches["system_price"]
            except KeyError:
                system_price_cache = data_source.caches["system_price"] = {}

            h_start = h["start-date"]
            rates = data_source.non_core_rate("system_price", h_start)[
                "gbp_per_nbp_mwh"
            ]

            try:
                try:
                    rdict = rates[key_format(h_start)]
                except KeyError:
                    rdict = rates[key_format(h_start - Timedelta(days=3))]
                sbp = float(rdict["sbp"] / 1000)
                ssp = float(rdict["ssp"] / 1000)
                system_price_cache[h_start] = (sbp, ssp)
            except KeyError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate cannot be found."
                )
            except TypeError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate 'rates_gbp_per_mwh' has the problem: "
                    f"{traceback.format_exc()}"
                )

        h["sbp"] = sbp
        h["sbp-gbp"] = h["nbp-kwh"] * sbp

        h["ssp"] = ssp
        h["ssp-gbp"] = h["nbp-kwh"] * ssp


def elexon_import(sess, log, set_progress, scripting_key):
    log("Starting to check System Prices.")
    contract_name = "system_price"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    contract_props = contract.make_properties()

    if not contract_props.get("enabled", True):
        log(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to True."
        )
        return

    for rscript in sess.scalars(
        select(RateScript)
        .where(RateScript.contract == contract)
        .order_by(RateScript.start_date.desc())
    ):
        ns = loads(rscript.script)
        rates = ns.get("gbp_per_nbp_mwh", {})
        if len(rates) == 0:
            fill_start = rscript.start_date
            break
        elif rates[key_format(rscript.finish_date)]["run"] == "DF":
            fill_start = rscript.finish_date + HH
            break
    res = download_file(log, scripting_key, "BESTVIEWPRICES_FILE")
    log(f"Extracting data from {hh_format(fill_start)}")

    sess.rollback()  # Avoid long-running transactions
    log(f"Received {res.status_code} {res.reason}")
    data = res.content
    book = xlrd.open_workbook(file_contents=data)
    sbp_sheet = book.sheet_by_index(1)
    ssp_sheet = book.sheet_by_index(2)

    sp_months = []
    sp_month = None
    for row_index in range(1, sbp_sheet.nrows):
        sbp_row = sbp_sheet.row(row_index)
        ssp_row = ssp_sheet.row(row_index)
        raw_date = datetime.datetime(
            *xlrd.xldate_as_tuple(sbp_row[0].value, book.datemode)
        )
        hh_date_ct = to_ct(raw_date)
        hh_date = to_utc(hh_date_ct)
        run_code = sbp_row[1].value
        for col_idx in range(2, 52):
            if hh_date >= fill_start:
                sbp_val = sbp_row[col_idx].value
                if sbp_val != "":
                    if hh_date.day == 1 and hh_date.hour == 0 and hh_date.minute == 0:
                        sp_month = {}
                        sp_months.append(sp_month)
                    ssp_val = ssp_row[col_idx].value
                    if sp_month is not None:
                        sp_month[hh_date] = {
                            "run": run_code,
                            "sbp": sbp_val,
                            "ssp": ssp_val,
                        }
            hh_date += HH
    log("Successfully extracted data.")
    last_date = sorted(sp_months[-1].keys())[-1]
    if last_date.month == (last_date + HH).month:
        del sp_months[-1]
    if "limit" in contract_props:
        sp_months = sp_months[0:1]
    for sp_month in sp_months:
        sorted_keys = sorted(sp_month.keys())
        month_start = sorted_keys[0]
        month_finish = sorted_keys[-1]
        rs = (
            sess.query(RateScript)
            .filter(
                RateScript.contract == contract,
                RateScript.start_date == month_start,
            )
            .first()
        )
        if rs is None:
            log(f"Adding a new rate script starting at {hh_format(month_start)}.")

            latest_rs = (
                sess.query(RateScript)
                .filter(RateScript.contract == contract)
                .order_by(RateScript.start_date.desc())
                .first()
            )

            contract.update_rate_script(
                sess,
                latest_rs,
                latest_rs.start_date,
                month_finish,
                loads(latest_rs.script),
            )
            rs = contract.insert_rate_script(sess, month_start, {})
            sess.flush()
        script = {
            "gbp_per_nbp_mwh": dict((key_format(k), v) for k, v in sp_month.items())
        }
        log(f"Updating rate script starting at {hh_format(month_start)}.")
        contract.update_rate_script(sess, rs, rs.start_date, rs.finish_date, script)
        sess.commit()



from collections import defaultdict
from decimal import Decimal

from chellow.e.bill_parsers.mm import (
    _handle_0100,
    _handle_0101,
    _handle_0460,
    _handle_0461,
    _handle_0860,
    _handle_1455,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_handle_0100():
    headers = {"issue_date": to_utc(ct_datetime(2023, 5, 2))}
    pre_record = (
        "5126507752002028004650                    "
        "002371905000002390596120850903650220  "
    )
    record = "".join(
        (
            "20240618",  # start_date=DATE_LENGTH
            "20240618",  # finish_date=DATE_LENGTH)
        )
    )
    _handle_0100(headers, pre_record, record)
    assert headers == {
        "issue_date": to_utc(ct_datetime(2023, 5, 2)),
        "account": "0023719050",
        "breakdown": {"vat": {}},
        "kwh": Decimal("0"),
        "reads": [],
        "reference": "000023905961",
    }


def test_handle_0101():
    headers = {}
    pre_record = ""
    record = "".join(
        (
            "20240618",  # start_date=DATE_LENGTH
            "20240618",  # finish_date=DATE_LENGTH)
        )
    )
    _handle_0101(headers, pre_record, record)
    assert headers == {
        "start_date": to_utc(ct_datetime(2024, 6, 18)),
        "finish_date": to_utc(ct_datetime(2024, 6, 18, 23, 30)),
    }


def test_handle_0460():
    headers = {"breakdown": defaultdict(int, {"vat": {}}), "reads": []}
    pre_record = ""
    record = "".join(
        (
            "xxxxxxxxxxxx",  # unknown_1=12
            "xxxxxxxxxxxx",  # unknown_2=12
            "10RS0050",  # code=8
            "000000000871",  # quantity=12
            "UNIT                  ",  # units=22
            "00000000000.0429",  # rate=16
            "20240401",  # unknown_date=DATE_LENGTH
            "000000761.59",  # gbp=12
            "                                   ",  # charge_description=35,
            "xxx",  # unknown_3=51
            "21",  # days=2
        )
    )
    _handle_0460(headers, pre_record, record)
    assert headers == {
        "reads": [],
        "breakdown": defaultdict(
            int,
            {
                "duos-red-gbp": Decimal("7.6159"),
                "duos-red-kwh": Decimal("871"),
                "duos-red-rate": {Decimal("0.0429")},
                "vat": {},
            },
        ),
    }


def test_handle_0461():
    headers = {"kwh": Decimal(0), "reads": []}
    pre_record = ""
    record = "".join(
        (
            "xxxxxxxxxxx",  # msn=11,
            "xx",  # unknown_1=2,
            "000000000539",  # prev_read_value=12,
            "000000000643",  # pres_read_value=12,
            "000001",  # coefficient=6,
            "   KWH",  # units=6,
            "000000000084",  # quantity=12,
            "UNIT  ",  # charge=6,
            "N",  # prev_read_type=1,
            "E",  # pres_read_type=1,
            "2276425654120",  # mpan_core=13,
            "11111111",  # mpan_top=8,
            "SINGLE             ",  # register_code=19,
            "20240401",  # pres_read_date=DATE_LENGTH,
            "20240301",  # prev_read_date=DATE_LENGTH,
        )
    )
    _handle_0461(headers, pre_record, record)
    assert headers == {
        "kwh": Decimal("84"),
        "reads": [
            {
                "msn": "xxxxxxxxxxx",
                "mpan": "11111111 22 7642 5654 120",
                "coefficient": Decimal("1"),
                "units": "kWh",
                "tpr_code": "00001",
                "prev_date": utc_datetime(2024, 3, 1, 0, 0),
                "prev_value": Decimal("539"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2024, 3, 31, 23, 0),
                "pres_value": Decimal("643"),
                "pres_type_code": "E",
            }
        ],
        "mpan_core": "22 7642 5654 120",
    }


def test_handle_0860():
    headers = {"breakdown": defaultdict(int, {})}
    pre_record = ""
    record = "".join(
        (
            "000000000456",  # metering_gbp=12
            "xxxxxxxxxxxx",  # unknown_1=12
            "xxxxxxxxxxxx",  # unknown_2=12
            "20240613",  # metering_date=DATE_LENGTH
            "x" * 80,  # description=80
        )
    )
    _handle_0860(headers, pre_record, record)
    assert headers == {
        "breakdown": {
            "metering-gbp": Decimal("4.56"),
        }
    }


def test_handle_1455():
    headers = {"breakdown": defaultdict(int, {})}
    pre_record = ""
    record = "".join(
        (
            "0000000000034",  # ccl_kwh=13
            "xxxxxxxxxx",  # unknown_1=8
            "0000000003.00",  # ccl_rate=15
            "000000000876",  # ccl_gbp=12
            "0       ",  # unknown_2=8
        )
    )
    _handle_1455(headers, pre_record, record)
    assert headers == {
        "breakdown": {
            "ccl-kwh": Decimal("34"),
            "ccl-rate": {Decimal("0.03")},
            "ccl-gbp": Decimal("8.76"),
        }
    }



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import HH, parse_mpan_core, to_ct, to_utc


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_date_ct(sheet, col, row):
    date_str = get_str(sheet, col, row)
    return to_ct(Datetime.strptime(date_str, "%d/%m/%Y"))


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


ELEM_LOOKUP = {
    "Admin Fee": {
        "gbp": "admin-gbp",
        "units": None,
        "rate": None,
    },
    "VAT": None,
    "GDUOS Charges": {
        "gbp": "duos-gbp",
        "units": None,
        "rate": None,
    },
    "Power Generation": {
        "gbp": "ssp-gbp",
        "units": "ssp-kwh",
        "rate": "ssp-rate",
    },
}


def _make_raw_bills(book):
    bills = {}
    mpan_lookup = {}
    sheet = book.worksheets[0]
    for row in range(2, len(sheet["A"]) + 1):
        val = get_cell(sheet, "A", row).value
        if val is None or val == "":
            break

        description = get_str(sheet, "T", row)
        desc_parts = [d.strip() for d in description.split("-")]
        if len(desc_parts) > 1:
            start_date_ct = to_ct(Datetime.strptime(desc_parts[1], "%b %y"))
            finish_date_ct = start_date_ct + relativedelta(months=1) - HH
        else:
            start_date_ct = get_date_ct(sheet, "M", row)
            finish_date_ct = get_date_ct(sheet, "N", row) + relativedelta(
                hours=23, minutes=30
            )

        desc_elem = desc_parts[0]
        start_date = to_utc(start_date_ct)
        finish_date = to_utc(finish_date_ct)
        reference = get_str(sheet, "G", row)
        issue_date = to_utc(get_date_ct(sheet, "I", row))

        bill_key = reference, start_date
        try:
            bill = bills[bill_key]
        except KeyError:
            bill = bills[bill_key] = {
                "bill_type_code": "N",
                "kwh": Decimal(0),
                "vat": Decimal("0.00"),
                "net": Decimal("0.00"),
                "gross": Decimal("0.00"),
                "reads": [],
                "breakdown": {},
                "issue_date": issue_date,
                "start_date": start_date,
                "finish_date": finish_date,
                "reference": reference,
            }

        net = round(get_dec(sheet, "V", row), 2)
        vat = round(get_dec(sheet, "W", row), 2)
        gross = round(get_dec(sheet, "X", row), 2)
        bill["net"] += net
        bill["vat"] += vat
        bill["gross"] += gross

        mpan_core_str = get_str(sheet, "E", row)
        if len(mpan_core_str) > 0:
            mpan_lookup[reference] = parse_mpan_core(mpan_core_str)

        mpan_core = mpan_lookup[reference]

        bill["mpan_core"] = mpan_core
        bill["account"] = mpan_core

        titles = ELEM_LOOKUP[desc_elem]
        if titles is None:
            continue

        breakdown = bill["breakdown"]
        breakdown[titles["gbp"]] = net

        units_title = titles["units"]
        rate_title = titles["rate"]
        units = get_dec(sheet, "P", row)
        rate = get_dec(sheet, "U", row) / Decimal(1000)

        if units_title is not None:
            breakdown[units_title] = units
        if rate_title is not None:
            breakdown[rate_title] = [rate]
    return bills.values()


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row = bills = None
        try:
            bills = _make_raw_bills(self.book)
        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills
